#ifndef PREC_B_H
#define PREC_B_H

#include "cudf.h"

#ifdef PREC_B
	// host
	#define MUBLAS_GEMM mublasBgemm
	#define MUBLAS_GEMV mublasBgemv
	#define MUBLAS_AXPY mublasBaxpy
	#define MUBLAS_SCAL mublasBscal
	#define MUBLAS_COPY mublasBcopy
	#define CUBLAS_GEMM	NULL
	#define CUBLAS_GEMV NULL
	#define CUBLAS_AXPY NULL
	#define CUBLAS_SCAL NULL
	#define CUBLAS_COPY NULL
	#define TYPE0 cudfreal
	#define ZERO0 make_cudfreal(0.,0.)
	#define ONE0 make_cudfreal(1.,0.)

	// vec1
	#define VEC1 1
	#define ZERO1 ZERO0
	#define TYPE1 TYPE0
	#define TYPEIN1 TYPE1
	#define ZEROIN1 ZERO1
	#define LOAD1 
	#define STORE1 
	#define ADD1 CUBADD
	#define MUL1 CUBMUL
	#define MAD1 CUBMAD

	#define MAD_1_111 MAD1
	#define MAD_1_111 MAD1
	#define MUL_1_11 MUL1

	// vec2
	#define VEC2 1
	#define TYPE2 cudfreal2
	#define ZERO2 make_cudfreal2(ZERO1,ZERO1)
	#define TYPEIN2 TYPE2
	#define ZEROIN2 ZERO2
	#define LOAD2 
	#define STORE2 
	#define ADD2 cuBadd2
	#define MUL12 cuBmul12
	#define MAD122 cuBmad122
	#define MAD212 cuBmad212
	#define MAD2 cuBmad2

	#define MAD_2_212 MAD212
	#define MAD_2_122 MAD122
	#define MUL_2_12 MUL12

	// vec4
	#define VEC4 0

	// DF-algorithm 
	#define CUBADD cuBaddV2
	#define CUBSUB cuBsubV2
	#define CUBMUL cuBmulV3
	#define CUBMAD cuBmuladd
#endif


// === Shuffle =====================================================
__device__ __forceinline__ void shuffleAdd (cudfreal &val, const int n) {
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
	val = CUBADD (val, make_cudfreal (__shfl_xor(val.x, n, T_W), __shfl_xor(val.y, n, T_W)));
#endif
}
// === Shuffle =====================================================


typedef float4 cudfreal2;

__device__ __forceinline__ cudfreal2
make_cudfreal2
	(const cudfreal x, const cudfreal y)
{
	cudfreal2 v;
	v.x = x.x;
	v.y = x.y;
	v.z = y.x;
	v.w = y.y;
	return v;
}

__host__ __device__ __forceinline__ bool operator== (const cudfreal a, const cudfreal b) {
	if ((a.x == b.x) && (a.y == b.y))
		return true;
	else
		return false;
}

__device__ __forceinline__ cudfreal
cuBmuladd
	(const cudfreal a, const cudfreal b, const cudfreal c)
{
	return CUBADD(CUBMUL(a, b), c);
}

__device__ __forceinline__ cudfreal2
cuBadd2
	(const cudfreal2 a, const cudfreal2 b)
{
	return make_cudfreal2 (CUBADD(make_cudfreal(a.x,a.y),make_cudfreal(b.x,b.y)), CUBADD(make_cudfreal(a.z,a.w),make_cudfreal(b.z,b.w)));
}

__device__ __forceinline__ cudfreal2
cuBmul12
	(const cudfreal a, const cudfreal2 b)
{
	return make_cudfreal2 (CUBMUL(a,make_cudfreal(b.x,b.y)), CUBMUL(a,make_cudfreal(b.z,b.w)));
}

__device__ __forceinline__ cudfreal2
cuBmad2
	(const cudfreal2 a, const cudfreal2 b, const cudfreal2 c)
{
	return make_cudfreal2 (CUBMAD(make_cudfreal(a.x,a.y),make_cudfreal(b.x,b.y),make_cudfreal(c.x,c.y)), CUBMAD(make_cudfreal(a.z,a.w),make_cudfreal(b.z,b.w),make_cudfreal(c.z,c.w)));
}

__device__ __forceinline__ cudfreal2
cuBmad122
	(const cudfreal a, const cudfreal2 b, const cudfreal2 c)
{
	return make_cudfreal2 (CUBMAD(a,make_cudfreal(b.x,b.y),make_cudfreal(c.x,c.y)), CUBMAD(a,make_cudfreal(b.z,b.w),make_cudfreal(c.z,c.w)));
}

__device__ __forceinline__ cudfreal2
cuBmad212
	(const cudfreal2 a, const cudfreal b, const cudfreal2 c)
{
	return make_cudfreal2 (CUBMAD(make_cudfreal(a.x,a.y),b,make_cudfreal(c.x,c.y)), CUBMAD(make_cudfreal(a.z,a.w),b,make_cudfreal(c.z,c.w)));
}

// === Texture =====================================================
texture<float2, cudaTextureType1D, cudaReadModeElementType> tex_mat_a;
texture<float2, cudaTextureType1D, cudaReadModeElementType> tex_mat_b;

static __forceinline__ __device__ cudfreal
tex_a(const int& i)
{
	float2 v = tex1Dfetch (tex_mat_a, i);
	return make_cudfreal (v.x, v.y);
}

static __forceinline__ __device__ cudfreal
tex_b(const int& i)
{
	float2 v = tex1Dfetch (tex_mat_b, i);
	return make_cudfreal (v.x, v.y);
}
// =================================================================

#endif
