#ifndef PREC_D_H
#define PREC_D_H

#ifdef PREC_D
	// host
	#define MUBLAS_GEMM mublasDgemm
	#define MUBLAS_GEMV mublasDgemv
	#define MUBLAS_AXPY mublasDaxpy
	#define MUBLAS_SCAL mublasDscal
	#define MUBLAS_COPY mublasDcopy
	#define CUBLAS_GEMM cublasDgemm
	#define CUBLAS_GEMV cublasDgemv
	#define CUBLAS_AXPY cublasDaxpy
	#define CUBLAS_SCAL cublasDscal
	#define CUBLAS_COPY cublasDcopy
	#define TYPE0 double
	#define ZERO0 0.
	#define ONE0 1.

	// vec1
	#define VEC1 1
	#define TYPE1 TYPE0
	#define ZERO1 ZERO0
	#define ADD1(A,B) ((A)+(B))
	#define MUL1(A,B) ((A)*(B))
	#define MAD1(A,B,C) (((A)*(B))+(C))

	#define TYPEIN1 TYPE1
	#define ZEROIN1 ZERO1
	#define LOAD1 
	#define STORE1 
	#define MAD_1_111 MAD1
	#define MAD_1_111 MAD1
	#define MUL_1_11 MUL1

	// vec2
	#define VEC2 1
	#define TYPE2 double2
	#define ZERO2 make_double2(0.,0.)
	#define ADD2 cuDadd2
	#define MUL12 cuDmul12
	#define MAD122 cuDmad122
	#define MAD212 cuDmad212
	#define MAD2 cuDmad2

	#define TYPEIN2 TYPE2
	#define ZEROIN2 ZERO2
	#define LOAD2 
	#define STORE2 
	#define MAD_2_212 MAD212
	#define MAD_2_122 MAD122
	#define MUL_2_12 MUL12

	// vec4
	#define VEC4 0
#endif

// === Shuffle =====================================================
__device__ __forceinline__ void shuffleAdd (double &val, const int n) {
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
	val += __hiloint2double(
		__shfl_xor(__double2hiint(val), n, T_W),
		__shfl_xor(__double2loint(val), n, T_W));
#endif
}
// === Shuffle =====================================================

// === Vector =====================================================
__device__ __forceinline__ double2
cuDadd2
	(const double2 a, const double2 b)
{
	return make_double2 (a.x+b.x, a.y+b.y);
}

__device__ __forceinline__ double2
cuDmul12
	(const double a, const double2 b)
{
	return make_double2 (a*b.x, a*b.y);
}

__device__ __forceinline__ double2
cuDmad2
	(const double2 a, const double2 b, const double2 c)
{
	return make_double2 (a.x*b.x+c.x, a.y*b.y+c.y);
}

__device__ __forceinline__ double2
cuDmad122
	(const double a, const double2 b, const double2 c)
{
	return make_double2 (a*b.x+c.x, a*b.y+c.y);
}

__device__ __forceinline__ double2
cuDmad212
	(const double2 a, const double b, const double2 c)
{
	return make_double2 (a.x*b+c.x, a.y*b+c.y);
}
// =================================================================

// === Texture =====================================================
texture<int2, cudaTextureType1D, cudaReadModeElementType> tex_mat_a;
texture<int2, cudaTextureType1D, cudaReadModeElementType> tex_mat_b;

static __forceinline__ __device__ double
tex_a(const int& i)
{
	int2 v = tex1Dfetch (tex_mat_a, i);
	return __hiloint2double (v.y, v.x);
}

static __forceinline__ __device__ double
tex_b(const int& i)
{
	int2 v = tex1Dfetch (tex_mat_b, i);
	return __hiloint2double (v.y, v.x);
}
// =================================================================

#endif
