#ifndef PREC_DB_H
#define PREC_DB_H

#include "cudf.h"

#ifdef PREC_DB
	// host
	#define MUBLAS_GEMM mublasDBgemm
	#define MUBLAS_GEMM_MAGMA mublasDBgemm_MAGMA
	#define MUBLAS_GEMV mublasDBgemv
	#define MUBLAS_AXPY mublasDBaxpy
	#define MUBLAS_SCAL mublasDBscal
	#define MUBLAS_COPY mublasDBcopy
	#define CUBLAS_GEMM	NULL
	#define CUBLAS_GEMV NULL
	#define CUBLAS_AXPY NULL
	#define CUBLAS_SCAL NULL
	#define CUBLAS_COPY NULL
	#define TYPE0 double
	#define ZERO0 0.
	#define ONE0 1.

	// vec1
	#define VEC1 1
	#define TYPE1 double
	#define TYPEIN1 cudfreal
	#define ZEROIN1 make_cudfreal(0.,0.)
	#define LOAD1 double_to_cudfreal
	#define STORE1 cudfreal_to_double
	#define MAD1 cuBmad
	#define MUL1 cuBmul

	// vec2
	#define VEC2 0//1
	#define TYPE2 double2
	#define ZEROIN2 make_cudfreal2(make_cudfreal(0.,0.),make_cudfreal(0.,0.))
	#define TYPEIN2 cudfreal2
	#define LOAD2 double2_to_cudfreal2
	#define STORE2 cudfreal2_to_double2
	#define MAD212 cuBmad_2_212 
	#define MAD122 cuBmad_2_122
	#define MUL12 cuBmul_2_12

	// vec4
	#define VEC4 0

	// DF-algorithm 
	// ADD 
	#define cuBadd cuBaddsloppy
//	#define cuBadd cuBaddieee
	// MUL
//	#define cuBmul cuBmulsloppy
//	#define cuBmul cuBmulieee
	#define cuBmul cuBmulhitachimod
	// MAD
	#define cuBmad cuBmuladd
	//#define cuBmad cuBmadfast
#endif

typedef float4 cudfreal2;

__device__ __forceinline__ cudfreal2
make_cudfreal2
	(const cudfreal x, const cudfreal y)
{
	cudfreal2 v;
	v.x = x.x;
	v.y = x.y;
	v.z = y.x;
	v.w = y.y;
	return v;
}

__host__ __device__ __forceinline__ bool operator== (const cudfreal a, const cudfreal b) {
	if ((a.x == b.x) && (a.y == b.y))
		return true;
	else
		return false;
}

__device__ __forceinline__ cudfreal
cuBmuladd
	(const cudfreal a, const cudfreal b, const cudfreal c)
{
	return cuBadd(cuBmul(a, b), c);
}

__device__ __forceinline__ cudfreal2
cuBadd2
	(const cudfreal2 a, const cudfreal2 b)
{
	return make_cudfreal2 (cuBadd(make_cudfreal(a.x,a.y),make_cudfreal(b.x,b.y)), cuBadd(make_cudfreal(a.z,a.w),make_cudfreal(b.z,b.w)));
}

__device__ __forceinline__ cudfreal2
cuBmul12
	(const cudfreal a, const cudfreal2 b)
{
	return make_cudfreal2 (cuBmul(a,make_cudfreal(b.x,b.y)), cuBmul(a,make_cudfreal(b.z,b.w)));
}

__device__ __forceinline__ cudfreal2
cuBmad212
	(const cudfreal2 a, const cudfreal b, const cudfreal2 c)
{
	return make_cudfreal2 (cuBmad(make_cudfreal(a.x,a.y),b,make_cudfreal(c.x,c.y)), cuBmad(make_cudfreal(a.z,a.w),b,make_cudfreal(c.z,c.w)));
}

__device__ __forceinline__ cudfreal2
cuBmad122
	(const cudfreal a, const cudfreal2 b, const cudfreal2 c)
{
	return make_cudfreal2 (cuBmad(a,make_cudfreal(b.x,b.y),make_cudfreal(c.x,c.y)), cuBmad(a,make_cudfreal(b.z,b.w),make_cudfreal(c.z,c.w)));
}

// double-dfreal mixed routines

__device__ __forceinline__ cudfreal2
double2_to_cudfreal2
	(const double2 d2)
{
	return make_cudfreal2 (double_to_cudfreal(d2.x), double_to_cudfreal(d2.y));
}


__device__ __forceinline__ double2
cudfreal2_to_double2
	(const cudfreal2 df2)
{
	return make_double2 (__dadd_rn ((double)df2.x, (double)df2.y), __dadd_rn ((double)df2.z, (double)df2.w));
}

__device__ __forceinline__ cudfreal2
cuBmul_2_12
	(const cudfreal Adf, const cudfreal2 Bdf2)
{
	return make_cudfreal2 (
				cuBmul(Adf, make_cudfreal(Bdf2.x, Bdf2.y)),
				cuBmul(Adf, make_cudfreal(Bdf2.z, Bdf2.w)));
}

__device__ __forceinline__ cudfreal2
cuBmad_2_212
	(const cudfreal2 Adf2, const cudfreal Bdf, const cudfreal2 Cdf2)
{
	return make_cudfreal2 (
				cuBmad(make_cudfreal(Adf2.x, Adf2.y), Bdf, make_cudfreal(Cdf2.x, Cdf2.y)),
				cuBmad(make_cudfreal(Adf2.z, Adf2.w), Bdf, make_cudfreal(Cdf2.z, Cdf2.w)));
} 

__device__ __forceinline__ cudfreal2
cuBmad_2_122
	(const cudfreal Adf, const cudfreal2 Bdf2, const cudfreal2 Cdf2)
{
	return make_cudfreal2 (
					cuBmad(Adf, make_cudfreal(Bdf2.x, Bdf2.y), make_cudfreal(Cdf2.x, Cdf2.y)),
					cuBmad(Adf, make_cudfreal(Bdf2.z, Bdf2.w), make_cudfreal(Cdf2.z, Cdf2.w)));
}

// === Texture =====================================================
texture<int2, cudaTextureType1D, cudaReadModeElementType> tex_mat_a;
texture<int2, cudaTextureType1D, cudaReadModeElementType> tex_mat_b;

static __forceinline__ __device__ double
tex_a(const int& i)
{
	int2 v = tex1Dfetch (tex_mat_a, i);
	return __hiloint2double (v.y, v.x);
}

static __forceinline__ __device__ double
tex_b(const int& i)
{
	int2 v = tex1Dfetch (tex_mat_b, i);
	return __hiloint2double (v.y, v.x);
}
// =================================================================

#endif
