#ifndef PREC_U_H
#define PREC_U_H

#include "cudd.h"

#ifdef PREC_U
	// host
	#define MUBLAS_GEMM mublasUgemm
	#define MUBLAS_GEMV mublasUgemv
	#define MUBLAS_AXPY mublasUaxpy
	#define MUBLAS_SCAL mublasUscal
	#define MUBLAS_COPY mublasUcopy
	#define CUBLAS_GEMM NULL
	#define CUBLAS_GEMV NULL
	#define CUBLAS_AXPY NULL
	#define CUBLAS_SCAL NULL
	#define CUBLAS_COPY NULL
	#define TYPE0 cuddcomplex
	#define ZERO0 make_cuddcomplex(0.,0.,0.,0.)
	#define ONE0 make_cuddcomplex(1.,0.,0.,0.)

	// vec1
	#define VEC1 1
	#define TYPE1 TYPE0
	#define ZERO1 ZERO0
	#define ADD1 cuUadd
	#define MUL1 cuUmul
	#define MAD1 cuUmad

	#define TYPEIN1 TYPE1
	#define ZEROIN1 ZERO1
	#define LOAD1 
	#define STORE1 
	#define MAD_1_111 MAD1
	#define MAD_1_111 MAD1
	#define MUL_1_11 MUL1

	// vec2
	#define VEC2 0

	// vec4
	#define VEC4 0

	// DD-algorithm 
	#define CUWADD cuWaddV2
	#define CUWSUB cuWsubV2
	#define CUWMUL cuWmulV3
	#define CUWMAD cuWmuladd
#endif

__host__ __device__ __forceinline__ bool operator== (const cuddreal a, const cuddreal b) {
	if ((a.x == b.x) && (a.y == b.y))
		return true;
	else
		return false;
}

__host__ __device__ __forceinline__ bool operator== (const cuddcomplex a, const cuddcomplex b) {
	if ((a.x == b.x) && (a.y == b.y))
		return true;
	else
		return false;
}

__device__ __forceinline__ cuddcomplex
cuUadd
	(const cuddcomplex a, const cuddcomplex b)
{
	cuddcomplex c;
	c.x = CUWADD (a.x, b.x);
	c.y = CUWADD (a.y, b.y);
	return c;
}

__device__ __forceinline__ cuddcomplex
cuUmul
	(const cuddcomplex a, const cuddcomplex b)
{
	cuddcomplex c;
	c.x = CUWSUB (CUWMUL (a.x, b.x), CUWMUL (a.y, b.y));
	c.y = CUWADD (CUWMUL (a.x, b.y), CUWMUL (a.y, b.x));
	return c;
}

__device__ __forceinline__ cuddcomplex
cuUmad
	(const cuddcomplex a, const cuddcomplex b, const cuddcomplex c)
{
	cuddcomplex d;
	d.x = CUWADD( CUWSUB (CUWMUL (a.x, b.x), CUWMUL (a.y, b.y)), c.x);
	d.y = CUWADD( CUWADD (CUWMUL (a.x, b.y), CUWMUL (a.y, b.x)), c.y);
	return d;
}

// === Shuffle =====================================================
__device__ __forceinline__ void shuffleAdd (cuddcomplex &val, const int n) {
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
	val = cuUadd (val, make_cuddcomplex ( 
		make_cuddreal(
		__hiloint2double( __shfl_xor(__double2hiint(val.x.x), n, T_W), __shfl_xor(__double2loint(val.x.x), n, T_W)),
		__hiloint2double( __shfl_xor(__double2hiint(val.x.y), n, T_W), __shfl_xor(__double2loint(val.x.y), n, T_W))),
		make_cuddreal(
		__hiloint2double( __shfl_xor(__double2hiint(val.y.x), n, T_W), __shfl_xor(__double2loint(val.y.x), n, T_W)),
		__hiloint2double( __shfl_xor(__double2hiint(val.y.y), n, T_W), __shfl_xor(__double2loint(val.y.y), n, T_W)))));
#endif
}
// === Shuffle =====================================================


#endif
